import functools
from abc import abstractmethod, ABC
from typing import Any
from weakref import WeakKeyDictionary

from .._OptionallyPresent import OptionallyPresent
from .._Absent import Absent
from ..._typing import RawJSONElement
from ...validator import JSONValidatorInstance


class Property(JSONValidatorInstance, ABC):
    """
    A property of a configuration describes a key to the object and the types
    of value it can take. The value of a property is a raw JSON element, or something
    that can be converted to/from raw JSON. The property name can be set to the
    empty string to inherit the attribute name it is set against in the
    configuration.
    """
    def __init__(self,
                 name: str = "",  # Default tells the property to inherit its attribute name
                 *,
                 optional: bool = False):
        self.__name: str = name
        self.__optional: bool = optional

        # The instance values of this property
        self.__values = WeakKeyDictionary()

    def name(self) -> str:
        """
        Gets the name of this property.

        :return:    The property name.
        """
        return self.__name

    def is_optional(self) -> bool:
        """
        Whether this property is an optional property.
        """
        return self.__optional

    def __get__(self, instance, owner):
        # If accessed from the class, return the property itself
        if instance is None:
            return self

        # Make sure a value has been set for this property
        if instance not in self.__values:
            raise AttributeError(f"{instance.__class__.__name__} has no value for property '{self.name()}'")

        # Get the value for this property
        return self.__values[instance]

    def get_as_raw_json(self, instance) -> OptionallyPresent[RawJSONElement]:
        """
        Gets the value of this property as raw JSON.

        :param instance:    The instance to get the value for.
        :return:            The raw JSON, or Absent.
        """
        # Make sure we have an instance
        self._require_instance(instance)

        # Get the value for the instance
        value = self.__get__(instance, None)

        # If the value is present, convert it to raw JSON
        if value is not Absent:
            value = self._value_as_raw_json(instance, value)

        return value

    def __set__(self, instance, value):
        # Must be accessed via an instance
        self._require_instance(instance)

        # Validate the value
        if value is not Absent:
            value = self.validate_value(instance, value)
        elif not self.is_optional():
            raise ValueError(f"Cannot set non-optional property '{self.name()}' as absent")

        # Set the value in the instance dictionary
        self._set_unchecked(instance, value)

    def _set_unchecked(self, instance, value):
        """
        Sets the value of this property without performing
        any validation.

        :param instance:    The instance to set the value against.
        :param value:       The value to set.
        """
        self.__values[instance] = value

    def __delete__(self, instance):
        # Can't delete the property itself
        if instance is None:
            raise AttributeError(f"Can't delete configuration properties")

        # Deleting the property value for an instance sets it as absent
        self.__set__(instance, Absent)

    def __set_name__(self, owner, name: str):
        # Can't have private properties
        if name.startswith("_"):
            raise AttributeError(f"Property names can't start with underscores ({name})")

        # Default the property name to the attribute name if one wasn't set
        if self.__name == "":
            self.__name = name

    @staticmethod
    def _require_instance(instance):
        """
        Helper method which raises an exception if instance is None.

        :param instance:    The instance.
        """
        # Make sure we are accessed via an instance
        if instance is None:
            raise AttributeError("Cannot access property through class (requires instance)")

    @functools.lru_cache(maxsize=None)
    def get_validator(self):
        return super().get_validator()

    @abstractmethod
    def _value_as_raw_json(self, instance, value) -> RawJSONElement:
        """
        Converts a value of this property to raw JSON.

        :param instance:    The instance the value belongs to.
        :param value:       The value to convert.
        :return:            The raw JSON.
        """
        pass

    @abstractmethod
    def validate_value(self, instance, value) -> Any:
        """
        Performs property value validation. Should raise a ValueError if
        validation fails, or return the actual value to store if validation
        passes.

        :param instance:    The instance the value is for.
        :param value:       The value to validate.
        :return:            The value to store.
        """
        pass



from typing import List, Iterable

from ...schema import one_of
from ._OfProperty import OfProperty
from ._Property import Property


class OneOfProperty(OfProperty):
    """
    Property which validates JSON that matches exactly one of
    a number of sub-properties.
    """
    def __init__(self,
                 name: str = "",
                 sub_properties: Iterable[Property] = tuple(),
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            sub_properties,
            one_of,
            optional=optional
        )

    def choose_subproperty(self, successes: List[bool]) -> int:
        # Start with an invalid index
        index = -1

        # Search the list
        for i, success in enumerate(successes):
            if success:
                # If a previous success occurred, we have matched multiple sub-properties
                if index >= 0:
                    raise ValueError(f"Value matched more than one sub-property")

                index = i

        if index == -1:
            raise ValueError(f"Value didn't match any sub-properties")

        return index



from abc import ABC
from typing import Type

from ....abc import is_abstract_class
from ..._typing import RawJSONElement
from ...schema import JSONSchema
from ...serialise import JSONValidatedBiserialisable
from ._Property import Property


class ProxyProperty(Property, ABC):
    """
    Property which takes values of a type that is serialisable
    to/from JSON.
    """
    def __init__(self,
                 name: str = "",
                 type: Type[JSONValidatedBiserialisable] = JSONValidatedBiserialisable,
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            optional=optional
        )

        # Type argument must be concrete
        if is_abstract_class(type):
            raise ValueError(f"type argument must be a concrete class, got {type.__name__}")

        # The type of values this property takes
        self.__type: Type[JSONValidatedBiserialisable] = type

    def type(self) -> Type[JSONValidatedBiserialisable]:
        """
        Gets the type of values this proxy-property takes.
        """
        return self.__type

    def get_json_validation_schema(self) -> JSONSchema:
        # Use the value-type's schema
        return self.__type.get_json_validation_schema()

    def _value_as_raw_json(self, instance, value: JSONValidatedBiserialisable) -> RawJSONElement:
        # Values are serialisable, so just serialise it
        return value.to_raw_json()

    def validate_value(self, instance, value):
        # Must be an instance of the correct type, or JSON deserialisable to the correct type
        if not isinstance(value, self.__type):
            try:
                if isinstance(value, str):
                    return self.__type.from_json_string(value)
                else:
                    return self.__type.from_raw_json(value)
            except Exception as e:
                raise ValueError(f"Error validating proxy value: {e}") from e

        return value



from ...schema import JSONSchema, TRIVIALLY_FAIL_SCHEMA
from ..._typing import RawJSONElement
from ._Property import Property


class RawProperty(Property):
    """
    Property which takes raw JSON values. Uses a schema for validation.
    """
    def __init__(self,
                 name: str = "",
                 schema: JSONSchema = TRIVIALLY_FAIL_SCHEMA,
                 *,
                 optional: bool = False):
        super().__init__(name, optional=optional)

        self.__schema = schema

    def get_json_validation_schema(self) -> JSONSchema:
        return self.__schema

    def _value_as_raw_json(self, instance, value) -> RawJSONElement:
        # Raw properties store raw JSON already, so just return it
        return value

    def validate_value(self, instance, value):
        # Validate the raw JSON with our schema
        try:
            self.validate_raw_json(value)
            return value
        except Exception as e:
            raise ValueError(f"{value} failed schema validation") from e



"""
Package for logging utilities.
"""
from ._standard_logger import (
    DEBUG_HANDLER_NAME,
    INFO_HANDLER_NAME,
    WARNING_HANDLER_NAME,
    ERROR_HANDLER_NAME,
    exact_level_filter,
    create_standard_handler,
    create_standard_logger
)



from abc import abstractmethod
from typing import Tuple, List, Any, Iterable, Callable
from weakref import WeakKeyDictionary

from ...schema import JSONSchema
from ..._typing import RawJSONElement
from .._OptionallyPresent import OptionallyPresent
from .._Absent import Absent
from ._Property import Property
from ._DummyInstance import DummyInstance


class OfProperty(Property):
    """
    Base class for OneOfProperty and AnyOfProperty.
    """
    def __init__(self,
                 name: str = "",
                 sub_properties: Iterable[Property] = tuple(),
                 schema_function: Callable[[Iterable[JSONSchema]], JSONSchema] = None,  # one_of/any_of/all_of
                 *,
                 optional: bool = False):
        # Consume the sub-properties
        sub_properties = tuple(sub_properties)

        # Must provide at least 2 sub-properties
        if len(sub_properties) < 2:
            raise ValueError(f"Of properties require at least 2 sub-properties")

        super().__init__(
            name,
            optional=optional
        )

        self.__sub_properties: Tuple[Property] = sub_properties
        self.__instance_keys = WeakKeyDictionary()
        self.__schema_function = schema_function

    def _get_current_subproperty(self, instance) -> Tuple[OptionallyPresent[Property],
                                                          OptionallyPresent[DummyInstance]]:
        """
        Gets the sub-property which contains the value for the
        given instance, and the key to that value.

        :return:    The property and the instance key to its value,
                    or a pair of Absents if the value is absent.
        """
        # Only valid for instances
        self._require_instance(instance)

        # Get the current value
        value = super().__get__(instance, None)

        # If it's an index, convert it to the sub-property at that index
        if isinstance(value, int):
            return self.__sub_properties[value], self.__instance_keys[instance]

        return Absent, Absent

    def get_json_validation_schema(self) -> JSONSchema:
        return self.__schema_function(
            *(
                property.get_json_validation_schema()
                for property in self.__sub_properties
            )
        )

    def _set_unchecked(self, instance, value):
        # If setting an instance's value to absent, discard its
        # current instance key if it has one
        if value is Absent and instance in self.__instance_keys:
            del self.__instance_keys[instance]

        # Set as usual
        super()._set_unchecked(instance, value)

    def select_from_valid_subproperties(self, instance, value) -> Tuple[Any, int]:
        """
        Tests the given value against all sub-properties, and returns
        the validated value along with the index of the sub-property
        it should be stored against.

        :param instance:    The instance the value is for.
        :param value:       The value.
        :return:            The value to store, and the sub-property index
                            to store it against.
        """
        # Find the properties this value is valid for
        values = []
        for property in self.__sub_properties:
            # Record the validated value, or Absent if validation failed
            try:
                validated_value = property.validate_value(instance, value)
            except Exception:
                validated_value = Absent

            values.append(validated_value)

        # Select the canonical index
        subproperty_selection = self.choose_subproperty([value is not Absent for value in values])

        # Make sure the selection is in range
        if not isinstance(subproperty_selection, int) or not (0 <= subproperty_selection < len(self.__sub_properties)):
            raise ValueError(f"Error in internal index for Of property. "
                             f"Index should be integer in [0:{len(self.__sub_properties)}), "
                             f"got {subproperty_selection}")

        # Get the selection
        value = values[subproperty_selection]

        # Make sure the selected value is one of the successful ones
        if value is Absent:
            raise ValueError(f"Error selecting sub-property: selected sub-property {subproperty_selection} "
                             f"which failed validation")

        return value, subproperty_selection

    @abstractmethod
    def choose_subproperty(self, successes: List[bool]) -> int:
        """
        Allows the sub-classes to choose which sub-property should
        store a value based on the success or failure of all sub-
        properties validating the value. Should raise an ValueError if it
        can't decide.

        :param successes:   The list of property validation successes/failures.
        :return:            The index of the property to regard as storing the
                            value.
        """
        pass

    def __get__(self, instance, owner):
        # No instance returns the property itself
        if instance is None:
            return self

        # Get the current sub-property
        subproperty, instance_key = self._get_current_subproperty(instance)

        # If the current sub-property is absent then so is the value
        if subproperty is Absent:
            return Absent

        return subproperty.__get__(instance_key, owner)

    def _value_as_raw_json(self, instance, value) -> RawJSONElement:
        # Get the sub-property the value is stored in
        subproperty, instance_key = self._get_current_subproperty(instance)

        # Use the sub-property to convert the value to raw JSON
        return subproperty._value_as_raw_json(instance_key, value)

    def validate_value(self, instance, value) -> Any:
        # Select the sub-property to use
        value, subproperty_index = self.select_from_valid_subproperties(instance, value)

        # Set the value against the selected sub-property
        key = DummyInstance()
        self.__sub_properties[subproperty_index].__set__(key, value)
        self.__instance_keys[instance] = key

        # Our stored value is the sub-property index
        return subproperty_index



from ._Property import Property
from ._ProxyProperty import ProxyProperty
from ._MapProxy import MapProxy
from ._RawProperty import RawProperty


class MapProperty(ProxyProperty):
    """
    Property which validates a map of strings to some other type.
    """
    def __init__(self,
                 name: str = "",
                 value_property: Property = RawProperty(),  # Default property always fails, so will need replacing
                 *,
                 optional: bool = False):
        # Create a closure class to proxy the map
        class ClosureMapProxy(MapProxy):
            @classmethod
            def value_property(cls) -> Property:
                return value_property

        super().__init__(
            name,
            ClosureMapProxy,
            optional=optional
        )

    def validate_value(self, instance, value):
        # Convert other map-proxies and raw dictionaries to our proxy-type
        if (isinstance(value, MapProxy) and not isinstance(value, self.type())) or \
           isinstance(value, dict):
            value = self.type()(value)

        return super().validate_value(instance, value)



from typing import List, Iterable

from ...schema import all_of
from ._OfProperty import OfProperty
from ._Property import Property


class AllOfProperty(OfProperty):
    """
    Property which validates JSON that matches all of
    a number of sub-properties.
    """
    def __init__(self,
                 name: str = "",
                 sub_properties: Iterable[Property] = tuple(),
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            sub_properties,
            all_of,
            optional=optional
        )

    def choose_subproperty(self, successes: List[bool]) -> int:
        # If all sub-properties match, just pick the first one
        if all(successes):
            return 0

        raise ValueError(f"Value didn't match all sub-properties")



from typing import List, Iterable

from ...schema import any_of
from ._OfProperty import OfProperty
from ._Property import Property


class AnyOfProperty(OfProperty):
    """
    Property which validates JSON that matches at least one of
    a number of sub-properties.
    """
    def __init__(self,
                 name: str = "",
                 sub_properties: Iterable[Property] = tuple(),
                 *,
                 optional: bool = False):
        super().__init__(
            name,
            sub_properties,
            any_of,
            optional=optional
        )

    def choose_subproperty(self, successes: List[bool]) -> int:
        # Search the list
        for i, success in enumerate(successes):
            # Return the first valid index found
            if success:
                return i

        raise ValueError(f"Value didn't match any sub-properties")
